# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import pathlib
from abc import abstractmethod

from vdk.api.job_input import IJobInput


class IStandaloneDataJob:
    """
    A contextmanager that initialises VDK and its plugins,
    triggers plugin lifecycle hooks needed to expose an initialised IJobInput on entry
    and finalization hooks when the context is exited

    NB: The following plugin hooks are not triggered:
        * CoreHookSpecs.vdk_command_line

    Sample usage::

        with StandaloneDataJobFactory.create(datajob_directory) as job_input:
            #... use job_input object to interact with underlying data systems.
    """

    @abstractmethod
    def __init__(
        self,
        data_job_directory: pathlib.Path | None,
        name: str | None = None,
        job_args: dict | None = None,
    ):
        """
        Initialises:
            * PluginRegistry
                * load_plugins_from_setuptools_entrypoints
                * core-plugins
                * ExecutionTrackingPlugin
            * Configuration
            * CoreContext
            * StateStore
        Triggers plugin lifecycle hooks:
            * CoreHookSpecs.vdk_start.call_historic
            * CoreHookSpecs.vdk_configure
            * CoreHookSpecs.vdk_initialize

        Arguments:
            data_job_directory: Optional[pathlib.Path]
                The source code of the data job that will be executed.
            name: Optional[str]
                The name of the job.  If omitted will be inferred from the director name.
            job_args: Optional[dict]
                Allows for users to pass arguments to data job run.
                Data Job arguments are also used for parameter substitution in queries, see execute_query docstring.
        """

    @abstractmethod
    def __enter__(self) -> IJobInput:
        """
        Executes on entry to the context block:
            * Adds a noop step DataJob hook to prevent any logic being dynamically executed from step files
            * Initialises the JobContext

        And then triggers plugin lifecycle hooks:
            * initialize_job
            * run_job

        Returns:
            An initialised IJobInput object
        """
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Executes on exit of the context block

        Triggers plugin lifecycle hooks:
            * finalize_job
            * vdk_exit
        """
        pass



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import List

from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.config.job_config import JobConfigKeys
from vdk.internal.builtin_plugins.config.vdk_config import LOG_CONFIG
from vdk.internal.builtin_plugins.notification import notification_base
from vdk.internal.builtin_plugins.notification import notification_configuration
from vdk.internal.builtin_plugins.notification.notification_configuration import (
    NotificationConfiguration,
)
from vdk.internal.builtin_plugins.notification.notification_configuration import (
    SmtpConfiguration,
)
from vdk.internal.builtin_plugins.run.execution_state import ExecutionStateStoreKeys
from vdk.internal.core import errors
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.context import CoreContext
from vdk.internal.core.statestore import CommonStoreKeys

log = logging.getLogger(__name__)


def __get_list(value: str) -> List[str]:
    result = value.split(";") if value else []
    result = [v.strip() for v in result if len(v.strip()) > 0]
    return result


def _notify(error_overall, user_error, configuration, state):
    notification_cfg = NotificationConfiguration(configuration)

    if (
        configuration.get_value(LOG_CONFIG).lower() == "local"
        and notification_cfg.get_notification_enabled() is False
    ):
        return
    log.debug(
        f"Prepare to send notification if necessary: error_overall: {error_overall}"
    )

    job_name = state.get(ExecutionStateStoreKeys.JOB_NAME)
    op_id = state.get(CommonStoreKeys.OP_ID)

    job_log_url_template = notification_cfg.get_job_log_url_pattern()
    cc = notification_cfg.get_email_cc_list()
    sender = notification_cfg.get_sender()
    smtp_cfg = SmtpConfiguration(configuration)

    if not error_overall:
        recipients = __get_list(
            configuration.get_value(JobConfigKeys.NOTIFIED_ON_JOB_SUCCESS.value)
        )
        if recipients:
            subject, body = notification_base.SuccessEmailNotificationMessageBuilder(
                job_name, job_log_url_template
            ).build(exec_type="run", op_id=op_id)
            notification_base.EmailNotification(
                recipients=recipients, cc=cc, smtp_cfg=smtp_cfg, sender=sender
            ).notify(subject, body)
        else:
            log.debug(
                "No recipients configured to send on successful run. No notification is sent"
            )
        return

    if user_error:
        recipients = __get_list(
            configuration.get_value(
                JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_USER_ERROR.value
            )
        )
        if recipients:
            subject, body = notification_base.UserErrorEmailNotificationMessageBuilder(
                job_name, job_log_url_template
            ).build(exec_type="run", op_id=op_id, msg=user_error)
            notification_base.EmailNotification(
                recipients=recipients, cc=cc, smtp_cfg=smtp_cfg, sender=sender
            ).notify(subject, body)
        else:
            log.debug(
                "No recipients configured to send on user error. No notification is sent"
            )
        return

    recipients = __get_list(
        configuration.get_value(
            JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_PLATFORM_ERROR.value
        )
    )

    if recipients:
        subject, body = notification_base.InfraErrorEmailNotificationMessageBuilder(
            job_name, job_log_url_template
        ).build(exec_type="run", op_id=op_id)
        notification_base.EmailNotification(
            recipients=recipients, cc=cc, smtp_cfg=smtp_cfg, sender=sender
        ).notify(subject, body)
    else:
        log.debug(
            "No recipients configured to send on platform error. No notification is sent"
        )

    log.debug("Notification for Data Job state has been sent to listed recipients")


class NotificationPlugin:
    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder):
        notification_configuration.add_definitions(config_builder)

    @hookimpl
    def vdk_exit(self, context: CoreContext, exit_code: int):
        if context.configuration.get_value(
            JobConfigKeys.ENABLE_ATTEMPT_NOTIFICATIONS.value
        ):
            _notify(
                errors.get_blamee_overall(),  # TODO: get this from context
                errors.get_blamee_overall_user_error(),  # TODO: get this from context
                context.configuration,
                context.state,
            )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
from dataclasses import dataclass
from datetime import datetime
from typing import Any
from typing import List
from typing import Optional

from vdk.internal.builtin_plugins.run.run_status import ExecutionStatus
from vdk.internal.core import errors
from vdk.internal.core.errors import ErrorMessage
from vdk.internal.core.errors import PlatformServiceError


@dataclass(frozen=True)
class StepResult:
    """
    The result of a step execution
    """

    # the name of the concrete step (e.g name of the file)
    name: str
    # TYPE_SQL, TYPE_PYTHON  or user-defined type
    type: str
    # when did the step started
    start_time: datetime
    # when did it finish
    end_time: datetime
    # status ?
    status: ExecutionStatus
    # string with details.
    details: Optional[str]
    # Exception if thrown
    exception: Optional[BaseException] = None
    # who is responsible for resolving the error
    blamee: Optional[errors.ResolvableBy] = None


class ExecutionResult:
    """
    Keep the execution result of a data job run.
    """

    def __init__(
        self,
        data_job_name: str,
        execution_id: str,
        start_time: datetime,
        end_time: datetime,
        status: ExecutionStatus,
        exception: Optional[BaseException],
        steps_list: List[StepResult],
    ):
        self.data_job_name = data_job_name
        self.execution_id = execution_id
        self.start_time = start_time
        self.end_time = end_time
        self.status = status
        self.steps_list = steps_list
        self.exception = exception

    def is_failed(self):
        return self.status == ExecutionStatus.ERROR

    def is_success(self):
        return self.status == ExecutionStatus.SUCCESS

    def get_exception(self):
        """
        Deprecated in favour of #get_exception_to_raise()
        """
        return self.get_exception_to_raise()

    def get_exception_to_raise(self):
        """
        Returns main exception to be used as a failure reason for the data job.
        """
        if self.exception:
            return self.exception
        step_exceptions = map(lambda x: x.exception, self.steps_list)
        step_exception = next(filter(lambda e: e is not None, step_exceptions), None)
        if step_exception:
            return step_exception
        else:
            return PlatformServiceError(
                ErrorMessage(
                    f"Data Job {self.data_job_name} failed",
                    "Data Job has failed",
                    "Failure is with unspecified reason. Seems like a bug in VDK.",
                    "Job will not complete",
                    "Retry the job. "
                    "Consider opening a ticket https://github.com/vmware/versatile-data-kit/issues",
                )
            )

    def __repr__(self):
        data = self.__dict__.copy()
        # make dates more human-readble
        data["start_time"] = self.start_time.isoformat()
        data["end_time"] = self.end_time.isoformat()
        if self.exception:
            data["exception_name"] = str(self.exception.__class__.__name__)

        step_lists_of_dicts = list()
        for step in self.steps_list:
            d = step.__dict__.copy()
            d["start_time"] = d["start_time"].isoformat()
            d["end_time"] = d["end_time"].isoformat()
            if step.exception:
                d["exception_name"] = str(step.exception.__class__.__name__)
                if step.exception.__cause__:
                    setattr(
                        d["exception"],
                        "cause_exception_name",
                        step.exception.__cause__.__class__.__name__,
                    )
                    setattr(
                        d["exception"],
                        "cause_exception",
                        step.exception.__cause__.__dict__,
                    )
            step_lists_of_dicts.append(d)

        data["steps_list"] = step_lists_of_dicts

        def default_serialization(o: Any) -> Any:
            return o.__dict__ if "__dict__" in dir(o) else str(o)

        return json.dumps(data, default=default_serialization, indent=2)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.job_properties import properties_config
from vdk.internal.core.config import ConfigurationBuilder


class PropertiesApiPlugin:
    """
    Define the basic configuration needed for Properties API.
    """

    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        properties_config.add_definitions(config_builder)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""This module contains the logic that
decides who is to blame, between Platform (SRE) Team and VDK Users,
when an exception occurs while executing a Data Job step.
"""
import os
import traceback
from pathlib import Path
from typing import Optional

from vdk.internal.core import errors


def whom_to_blame(exception, executor_module, data_job_path: Optional[Path] = None):
    """
    :param exception: Exception object that has led to Data Job failure.
    :param executor_module: name of module that executes User Code.
    :param data_job_path: path object of the data job directory.
    :return: ResolvableBy.PLATFORM_ERROR if exception was recognized as Platform Team responsibility.
             errors.ResolvableBy.USER_ERROR if exception was recognized as User Error.
    """
    if isinstance(exception, errors.BaseVdkError):
        return errors.find_whom_to_blame_from_exception(exception)
    if is_user_error(exception, data_job_path):
        return errors.ResolvableBy.USER_ERROR
    if _is_exception_from_vdk_code(
        exception, executor_module
    ):  # We want to avoid blaming users for errors raised from VDK code
        return (
            errors.ResolvableBy.PLATFORM_ERROR
        )  # Because we can't be 100% sure if they are user errors
    else:
        return errors.ResolvableBy.USER_ERROR


def _is_exception_from_vdk_code(exception, executor_module):
    executor_module = os.path.abspath(executor_module)
    vdk_code_directory = os.path.dirname(executor_module)
    call_list = traceback.format_tb(exception.__traceback__)
    call_list.reverse()  # First element is most recent call
    if len(call_list) == 0:
        return True

    for call in call_list:
        caller_module = call.split('"')[1]  # Extract module path from stacktrace call.
        if vdk_code_directory in caller_module and caller_module != executor_module:
            return True
        elif (
            caller_module == executor_module
        ):  # User code starts from this module always.
            break

    return False


def is_user_error(
    received_exception: Exception, data_job_path: Optional[Path] = None
) -> bool:
    """
    Returns if exception should be user error
    """
    return (
        _is_file_size_limit_exceeded(received_exception)
        or _is_new_thread_error(received_exception)
        or _is_timeout_error(received_exception)
        or _is_memory_limit_exceeded(received_exception)
        or _is_direct_user_code_error(received_exception, job_path=data_job_path)
    )


def _is_memory_limit_exceeded(exception):
    return errors.exception_matches(
        exception,
        classname_with_package="MemoryError",
        exception_message_matcher_regex=".*",
    )


def _is_file_size_limit_exceeded(exception):
    return errors.exception_matches(
        exception,
        classname_with_package="OSError",
        exception_message_matcher_regex=".*File size limit.*",
    )


def _is_new_thread_error(exception):
    return errors.exception_matches(
        exception,
        classname_with_package="RuntimeError",
        exception_message_matcher_regex=".*new thread.*",
    )


def _is_timeout_error(exception):
    return errors.exception_matches(
        exception,
        classname_with_package=".*TimeoutError.*",
        exception_message_matcher_regex=".*Duration of Data Job exceeded.*seconds.*",
    )


def _is_direct_user_code_error(exception: Exception, job_path: Optional[Path]):
    data_job_path = str(job_path) if job_path else None
    if not data_job_path:
        return False

    # Get exception traceback as a list
    call_list = traceback.format_tb(exception.__traceback__)
    if len(call_list) == 0:
        return False

    last_call = call_list[-1]
    last_caller_module = last_call.split('"')[1]

    # Check if the data job path is contained in the last exception call from the exception
    # traceback. If it is, it is safe to assume that the exception was generated directly in
    # user code, and not somewhere else in the vdk code or one of vdk's plugins.
    return True if data_job_path in last_caller_module else False



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import Any
from typing import cast

from vdk.api.plugin.connection_hook_spec import ConnectionHookSpec
from vdk.api.plugin.hook_markers import hookimpl
from vdk.api.plugin.plugin_registry import IPluginRegistry
from vdk.internal.builtin_plugins.connection.execution_cursor import (
    ExecuteOperationResult,
)
from vdk.internal.builtin_plugins.connection.execution_cursor import ExecutionCursor
from vdk.internal.core.errors import ErrorMessage
from vdk.internal.core.errors import PlatformServiceError


class DefaultConnectionHookImpl:
    """
    The default implementation of execute operation.
    Generally it should not be overridden.
    To "override" with new implementation, make a new hook which returns non-None ExecuteOperationResult.
    See ConnectionHookSpec documentation for more details.
    """

    @hookimpl(trylast=True)
    def db_connection_execute_operation(self, execution_cursor: ExecutionCursor) -> Any:
        managed_operation = execution_cursor.get_managed_operation()
        native_cursor = execution_cursor.get_native_cursor()
        if managed_operation.get_parameters():
            native_result = native_cursor.execute(
                managed_operation.get_operation(), managed_operation.get_parameters()
            )
        else:
            native_result = native_cursor.execute(managed_operation.get_operation())
        return ExecuteOperationResult(native_result)


class ConnectionHookSpecFactory:
    """
    Class used to create properly initialized ConnectionHookSpec instance to use to execute the underlying hooks
    """

    def __init__(self, plugin_registry: IPluginRegistry):
        self.__plugin_registry = plugin_registry

    def get_connection_hook_spec(self) -> ConnectionHookSpec:
        """
        Returns ConnectionHookSpec class which would act as a relay and invoke the underlying implemented hooks
        It's generally a ConnectionHookSpec cast of plugin_registry.PluginHookRelay to enable easier hook invocations.
        It also initializes some of the hooks (now only db_connection_execute_operation) with default implementations.
        :return: ConnectionHookSpec
        """
        if self.__plugin_registry:
            if not self.__plugin_registry.has_plugin(
                DefaultConnectionHookImpl.__name__
            ):
                self.__plugin_registry.load_plugin_with_hooks_impl(
                    DefaultConnectionHookImpl(), DefaultConnectionHookImpl.__name__
                )
            return cast(ConnectionHookSpec, self.__plugin_registry.hook())
        else:
            raise PlatformServiceError(
                ErrorMessage(
                    "Managed Cursor not initialized properly",
                    "Cannot connect to database using vdk managed cursor",
                    "Plugin registry is not initialized. That seems like a bug.",
                    "Without plugin registry the connection cannot be started",
                    "Open a vdk github issue "
                    "and/or revert to previous version of vdk-core.",
                )
            )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pathlib
from abc import ABC
from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import NewType
from typing import Optional
from typing import Tuple
from typing import Union

from vdk.internal.builtin_plugins.connection.managed_connection_base import (
    ManagedConnectionBase,
)

# represents any compatible PEP249 Connection class (most of python database clients)
PEP249Connection = Any


class IPropertiesServiceClient:
    """
    Client interface for reading and writing job properties.
    Different backend services can implement it so properties can be stored in different places
    (e.g databae, inmemory, vault)
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def read_properties(self, job_name: str, team_name: str) -> Dict:
        """
        Read properties from the backend - returns a dictionary of all properties for the given data job.
        :param team_name: The name of the team the job belongs to
        :param job_name: the name of the job
        """
        pass

    @abstractmethod
    def write_properties(self, job_name: str, team_name: str, properties: Dict) -> Dict:
        """
        Write properties to the backend service. It overwrites all properties completely.
        :param team_name: The name of the team the job belongs to
        :param job_name: the name of the job
        :param properties: dictionary with the new job's properties
        :return properties: dictionary with the properties handled
        """
        pass


IPropertiesFactory = Callable[[], IPropertiesServiceClient]


class IPropertiesRegistry(ABC):
    """
    Properties registry to enable register implementation Properties service
    """

    @abstractmethod
    def set_properties_factory_method(
        self,
        properties_type: str,
        properties_factory: IPropertiesFactory,
    ) -> None:
        """
        Register properties implementation backend.
        Properties API enable keeping state and secrets of a data job.
        Default implementation is in-memory so it's strongly recommended to install vdk-properties plugin
        which provides API based properties implementation

        IPropertiesServiceClient is used as provides logic of how properties are persisted
        and is part of implementation of IProperties interface.
        See

        :param properties_type: str
                string identifying the type of properties being set.
                This enable users to configure what type of properties implemenation to use.
        :param properties_factory: Callable[[], IPropertiesServiceClient]
               method or interface that returns correctly initialized implementation of IPropertiesServiceClient interface
        """
        pass


class ITemplateRegistry:
    """
    Template registry interface for adding new data jobs as templates.

    see job_input.ITemplate for how the template will be used by Data Engineers.

    Templates functionally enables to compose data jobs within themselves.
    For example common platform team can define common data jobs as templates and use them within other data jobs.
    This interface is used by plugin developers to register new templates.

    """

    # TODO: add supported arguments so that they can be validated
    @abstractmethod
    def add_template(self, name: str, template_directory: pathlib.Path):
        """
        Register a new template.
        You usually would want to register template with job_initialize plugin hook.

        For example:

        .. code-block:: python

            @hookimpl
            def initialize_job(self, context: JobContext) -> None:
                context.templates.add_template(
                    "template_name", pathlib.Path(template_name.job_module.__file__)
                )

        :param name: The name of the template . This name will be used by users of vdk in their data jobs when invoking it
        :param template_directory: the directory which contains the implementation of the template. This is a data job which accept arguments.
        VDK will automatically run the data job internally and return result (success or not to the user.). See ITemplate for more details.
        :return:
        """
        pass


class IManagedConnectionRegistry:
    def add_open_connection_factory_method(
        self,
        dbtype: str,
        open_connection_func: Callable[
            [], Union[ManagedConnectionBase, PEP249Connection]
        ],
    ) -> None:
        """

        Register new connection factory method.
        By registering the connection it will be made available by data engineers using job_input.

        For example:

        .. code-block:: python

            @hookimpl(tryfirst=True)
            def vdk_configure(config_builder: ConfigurationBuilder) -> None:
                # Here we define what configuration settings are needed to initialize the connection
                config_builder.add(
                    key="host",  description="The host we need to connect ."
                )
                config_builder.add(
                    key="port", description="The port to connect to"
                )

            @hookimpl
            def initialize_job(context: JobContext) -> None:
                context.connections.add_open_connection_factory_method(
                    "BIG_QUERY", lambda: dbapi.connect(host=cfg.get_value('host'), post=cfg.get_value('port')
                )

        :param dbtype: the name of the database connection - e.g: redshift, impala, presto, big-query, postgres, etc.
        :param open_connection_func:
            The function must implement the code necessary to open a new connection that includes getting all configuration necessary like url, user, etc.
            The function can return either original PEP 249 (DBAPI) Connection from the python library for the given database
            or can wrap the connection into ManagedConnection by subclassing . The second method is only necessary
            if plugin does some pre / post processing and want to decorate some methods (e.g to collecto stats for each query)

        :return:
        """
        pass


class IIngesterPlugin:
    """
    Interface representing the APIs that can be used when creating Ingester
    Plugins. Plugins implement one of the methods, depending on their needs.
    If more complex processing is required, plugins can be chained
    together.

    The lifecycles of the three methods are as follows:

    ingest_payload() - plugins implementing this method are intended
                       to do actual data ingestion. Versatile Data Kit will
                       call this method when the payload that data engineers
                       have passed is ready to be ingested.

    pre_ingest_process() - plugins implementing this method do not aim at
                        ingesting the payload, rather to do some processing
                        of the payload before it is ingested.

    post_ingest_process() - plugins implementing this method aim at doing
                            some metadata processing after the payload is
                            ingested. The main use-case is telemetry data
                            collection.
                            NOTE: This method can be used only by plugins
                            that implement ingest_payload().

    Example plugin chaining workflows:

    Workflow 1
    ----------
    _____________________      _________________________      ______________
    |      Data Job     |      |    vdk-ingest-http    |      |            |
    |-------------------|      |-----------------------|      | Data Lake/ |
    |                   |      |                       |      | Warehouse/ |
    |send_object_       | ---> |   ingest_payload()    | ---> | Third-party|
    |    for_ingestion()|      |                       |      | endpoint   |
    |___________________|      |_______________________|      |____________|


    Workflow 2
    ----------                                                  ______________
                                                                |            |
    _____________________      ___________________________      | Data Lake/ |
    |      Data Job     |      |     vdk-ingest-http     |      | Warehouse/ |
    |-------------------|      |-------------------------|      | Third-party|
    |                   |      | -- ingest_payload()     | ---> | endpoint   |
    |send_object_       | ---> | |                       |      |____________|
    |    for_ingestion()|      | |                       |      ______________
    |                   |      | -->post_ingest_process()| ---> |            |
    |___________________|      |_________________________|      | Telemetry  |
                                                                | Endpoint   |
                                                                |____________|

    Workflow 3
    ----------
    _____________________   _______________   _____________   ______________
    |      Data Job     |   |vdk-validate-|   |vdk-ingest-|   |            |
    |                   |   |    data     |   |  http     |   |            |
    |-------------------|   |-------------|   |-----------|   | Data Lake/ |
    |                   |   |             |   |           |   | Warehouse/ |
    |send_object_       |-->|pre_ingest   |-->|ingest     |-->| Third-party|
    |    for_ingestion()|   |   _process()|   | _payload()|   | endpoint   |
    |                   |   |             |   |           |   |            |
    |___________________|   |_____________|   |___________|   |____________|
    """

    IngestionMetadata = NewType("IngestionMetadata", Dict)

    def ingest_payload(
        self,
        payload: List[dict],
        destination_table: Optional[str],
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
        metadata: Optional[IngestionMetadata] = None,
    ) -> Optional[IngestionMetadata]:
        """
        Do the actual ingestion of the payload

        :param payload: List[dict]
            The payload to be ingested, split into 1 or many dictionary
            objects.
            Note: The memory size of the list is dependent on the
            payload_size_bytes_threshold attribute.
        :param destination_table: Optional[string]
            (Optional) The name of the table, where the data should be
            ingested into. This argument is optional, and needs to be
            considered only when the name of the destination table is not
            included in the payload itself.
        :param target: Optional[string]
            (Optional) Used to identify where the data should be ingested into.
                Specifies a data source and its destination database.
                The values for this parameter can be in the format
                `<some-data-source_and-db-table>`, or as a URL.
                Example:
                    http://example.com/<some-api>/<data-source_and-db-table>
            This parameter does not need to be used, in case the
            `INGEST_TARGET_DEFAULT` configuration variable is set. This can be
            made by plugins, which may set default value, or it can be
            overwritten by users.
        :param collection_id: string
            (Optional) An identifier to indicate that data from different
            method invocations belong to same collection. Defaults to
            "data_job_name|OpID", meaning all method invocations from a data
            job run will belong to the same collection.
        :param metadata: Optional[IngestionMetadata] dictionary object
            containing metadata produced and possibly used by pre-ingest,
            ingest, or post-ingest plugins.
            NOTE: A read-only parameter. Whatever modifications are done to
            this object, once returned, it is treated as a new object.

        :return: [Optional] IngestionMetadata, containing data about the
        ingestion process (information about the result from the ingestion
        operation, plugin-populated ingestion metadata, etc.). This
        metadata could be used for post-ingestion processing operations like
        sending telemetry data about the ingested payload.

        .. code-block:: python
            def ingest_payload(self,
                               payload: List[dict],
                               destination_table: Optional[str],
                               target: Optional[str] = None,
                               collection_id: Optional[str] = None,
            ) -> Optional[IngestionMetadata]:
                result = None
                metadata: IngestionMetadata = dict()
                metadata["destination_table"] = destination_table
                metadata["target"] = target

                try:
                    result = requests.post('example.com', data=payload)
                except:
                    log.info("Payload sent")

                if result:
                    metadata["http_status"] = result.status()

                return metadata

        :exception: Exceptions raised while ingesting data can either be
            handled at plugin level, or they will be automatically caught at
            the vdk-core level.
        """
        pass

    def pre_ingest_process(
        self,
        payload: List[dict],
        destination_table: Optional[str] = None,
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
        metadata: Optional[IngestionMetadata] = None,
    ) -> Tuple[List[Dict], Optional[IngestionMetadata]]:
        """
        Do some processing on the ingestion payload before passing it to the
        actual ingestion.

        For example, if we want to ensure that all values in the payload are
        string before the actual ingestion happens, we could have a plugin,
        implementing this method, and doing some payload processing:

        .. code-block:: python
            def pre_ingest_process(self,
                payload: List[dict],
                destination_table: Optional[str],
                target: Optional[str],
                collection_id: Optional[str],
                metadata: Optional[IngestionMetadata],
            ) -> Tuple[List[Dict], Optional[IngestionMetadata]]:
                # Ensure all values in the payload are strings
                processed_payload = \
                        [{k: str(v) for (k,v) in i.items()} for i in payload]

                return processed_payload, metadata

        :param payload: List[dict] the ingestion payload to be processed.
            NOTE: A read-only parameter. Whatever modifications are done to
            this object, once returned, it is treated as a new object.
        :param destination_table: Optional[string]
            (Optional) The name of the table, where the data should be
            ingested into. This argument is optional, and needs to be
            considered only when the name of the destination table is not
            included in the payload itself.
            NOTE: A read-only parameter. It is not expected to be modified and
            returned.
        :param target: Optional[string]
            (Optional) Used to identify where the data should be ingested into.
                Specifies a data source and its destination database.
                The values for this parameter can be in the format
                `<some-data-source_and-db-table>`, or as a URL.
                Example:
                      http://example.com/<some-api>/<data-source_and-db-table>
            This parameter does not need to be used, in case the
            `INGEST_TARGET_DEFAULT` configuration variable is set. This can be
            made by plugins, which may set default value, or it can be
            overwritten by users.
            NOTE: A read-only parameter. It is not expected to be modified and
            returned.
        :param collection_id: string
            (Optional) An identifier to indicate that data from different
            method invocations belong to same collection. Defaults to
            "data_job_name|OpID", meaning all method invocations from a data
            job run will belong to the same collection.
            NOTE: A read-only parameter. It is not expected to be modified and
            returned.
        :param metadata: Optional[IngestionMetadata], a dictionary object
            containing metadata produced by the pre-ingest plugin,
            and possibly used by other pre-ingest, ingest, or post-ingest
            plugins.
            NOTE: A read-only parameter. Whatever modifications are done to
            this object, once returned, it is treated as a new object.
        :return: Tuple[List[Dict], Optional[IngestionMetadata]], a tuple
            containing the processed payload objects and an
            IngestionMetadata object with ingestion metadata information.
            If no metadata is being added or processed, the metadata object
            passed to this method can be returned as is (see code block above).

        :exception: If an exception occurs during this operation, all other
            pre-processing operations and the ingestion of the payload would
            be terminated to prevent data corruption. In case the ingestion is
            expected to happen regardless if there are exceptions or not,
            then the exception handling needs to be implemented within this
            method.
        """
        pass

    def post_ingest_process(
        self,
        payload: Optional[List[dict]] = None,
        destination_table: Optional[str] = None,
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
        metadata: Optional[IngestionMetadata] = None,
        exception: Optional[Exception] = None,
    ) -> Optional[IngestionMetadata]:
        """
        Do post-ingestion processing of the ingestion payload
        or metadata from the ingestion operation.

        Example Implementation for telemetry:

        .. code-block:: python

            def post_ingest_process(
                self,
                payload: Optional[List[dict]],
                destination_table: Optional[str],
                target: Optional[str],
                collection_id: Optional[str],
                metadata: Optional[IngestionMetadata],
                exception: Optional[Exception],
            ) -> Optional[IngestionMetadata]:

                # Prepare telemetry
                telemetry = {}
                payload_sizes = [sys.getsizeof(i) for i in payload]
                telemetry['payload_size'] = sum(payload_sizes)
                telemetry['caught_ingest_exceptions'] = exceptions
                telemetry |= metadata

                # Send telemetry to wherever is needed.
                result = requests.post('example.com', data=telemetry)
                telemetry["telemetry_req_status"] = result.status_code

                return telemetry


        :param payload: Optional[List[dict]]
            The payload that has been ingested, represented as 0 or many
            dictionary objects.
            NOTE: A read-only parameter. It is not expected to be modified and
            returned.
        :param destination_table: Optional[string]
            (Optional) The name of the table, where the data should be
            ingested into. This argument is optional, and needs to be
            considered only when the name of the destination table is not
            included in the payload itself.
            NOTE: A read-only parameter. It is not expected to be modified and
            returned.
        :param target: Optional[string]
            (Optional) Used to identify where the data should be ingested into.
                Specifies a data source and its destination database.
                The values for this parameter can be in the format
                `<some-data-source_and-db-table>`, or as a URL.
                Example:
                      http://example.com/<some-api>/<data-source_and-db-table>
            This parameter does not need to be used, in case the
            `INGEST_TARGET_DEFAULT` configuration variable is set. This can be
            made by plugins, which may set default value, or it can be
            overwritten by users.
            NOTE: A read-only parameter. It is not expected to be modified and
            returned.
        :param collection_id: string
            (Optional) An identifier to indicate that data from different
            method invocations belong to same collection. Defaults to
            "data_job_name|OpID", meaning all method invocations from a data
            job run will belong to the same collection.
            NOTE: A read-only parameter. It is not expected to be modified and
            returned.
        :param metadata: Optional[IngestionMetadata]
            The metadata from the ingestion operation, that plugin developers
            have decided to process further. This metadata can be either an
            IngestionMetadata object (which is effectively a dictionary), or
            None.
            NOTE: A read-only parameter. Whatever modifications are done to
            this object, once returned, it is treated as a new object.
        :param exception: Optional[Exception]
            A caught exception (if any) encountered while ingesting data.
            NOTE: A read-only parameter. Whatever modifications are done to
            this object, once returned, it is treated as a new object.
        :return: Optional[IngestionMetadata], an IngestionMetadata object
            with information about this and possibly the previous processes (
            pre-ingestion, ingestion, post-ingestion).
        :exception: If an exception occurs during this operation, all other
            post-processing operations would be terminated to prevent data
            corruption. In case all post-process operations are expected to
            be completed regardless if there are exceptions or not, then the
            exception handling needs to be implemented within this method.
        """
        pass


class IIngesterRegistry:
    """
    Ingester Registry interface for adding different types of ingester plugins.
    """

    def add_ingester_factory_method(
        self,
        method: str,
        ingester_plugin: Callable[[], IIngesterPlugin],
    ) -> None:
        """
        Register new ingester objects.
        By registering ingester objects, data engineers would be able to use the
        same vdk setup with multiple ingester plugins, and direct where the data
        should be ingested by only specifying the `method` when calling
        send_object_for_ingestion()/send_tabular_data_for_ingestion().

        :param method: str
               Ingestion method to be associated with a plugin.
               For example:
                    method='file' -> for ingest to file plugin
                    method='http' -> for ingest over http plugin
                    method='kafka' -> for ingest to kafka endpoint plugin
               The value for the method parameter will be converted to lowercase and
               stored in this way with the aim of preventing issues where a user has
               set a variable to the correct string for a particular method, but the
               case is wrong.
        :param ingester_plugin: Callable[[], IIngesterPlugin]
               The ingester must implement the ingestion mechanism indicated by the
               method parameter.
        """
        pass



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
    Setup file.
    Use setup.cfg to configure your project.
"""
import sys

from pkg_resources import require
from pkg_resources import VersionConflict
from setuptools import setup

try:
    require("setuptools>=38.3")
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)


if __name__ == "__main__":
    setup()




RELEASE_VERSION='0.3.568835525'
BUILD_DATE='Tue Jun 21 09:38:28 UTC 2022'
BUILD_MACHINE_INFO='Linux runner-opyxurxq-project-28359933-concurrent-0sx4g2 5.4.181-99.354.amzn2.x86_64 #1 SMP Wed Mar 2 18:50:46 UTC 2022 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='2617785041'
GIT_COMMIT_SHA='4c1a5801fc326011d14da2314e04da054cebf8e1'
GIT_BRANCH='main'
